import json
import shutil
from itertools import repeat
from multiprocessing.pool import Pool
from os import cpu_count, makedirs
from os.path import join, exists

import requests
import vk_api

API_VERSION = '5.92'
APP_ID = 6781880

users = {}


def auth_handler():
    key = input('Введите код двухфакторной аутентификации: ')
    remember_device = True
    return key, remember_device


def captcha_handler(captcha):
    key = input(f"Введите капчу ({captcha.get_url()}): ").strip()
    return captcha.try_again(key)


def login_vk(token=None, login=None, password=None):
    try:
        if token:
            vk_session = vk_api.VkApi(token=token, app_id=APP_ID, auth_handler=auth_handler, api_version=API_VERSION)
        elif login and password:
            vk_session = vk_api.VkApi(login, password, captcha_handler=captcha_handler, app_id=APP_ID,
                                      api_version=API_VERSION, auth_handler=auth_handler)
            vk_session.auth(token_only=True, reauth=True)
        else:
            raise KeyError('Не введен токен или пара логин-пароль')
        return vk_session.get_api()
    except Exception as e:
        print(e)


def get_albums(vk, vk_tools):
    albums = []
    for album in vk.photos.getAlbums(need_system=1)['items']:
        photos = vk_tools.get_all(values={'album_id': album['id'], 'photo_sizes': 1}, method='photos.get',
                                  max_count=1000)
        albums.append({'name': album['title'], 'photos': [p['sizes'][-1]['url'] for p in photos['items']]})
    return albums


def get_docs(vk):
    doc_list = vk.docs.get()
    docs = {'count': doc_list['count'], 'size': 0, 'docs': []}
    for doc in doc_list['items']:
        docs['docs'].append({'url': doc['url'], 'name': f"{doc['title']}_{doc['id']}", 'ext': doc['ext']})
        docs['size'] += doc['size']
    return docs


def get_conversations(vk, vk_tools):
    # получаем список диалогов
    conversation_list = vk_tools.get_all(method='messages.getConversations', max_count=100,
                                         values={'extended': 1, 'fields': 'first_name, last_name, name'})
    conversations = []
    # формируем список обработанных диалогов
    for conversation in conversation_list['items']:
        conversations.append(get_dialog_data(conversation, vk, vk_tools))
    return conversations


# сборка нужных данных в одной структуре
def get_dialog_data(conversation, vk, vk_tool):
    # запоминаем id, type(chat, user, group), name
    dialog = {'id': conversation['conversation']['peer']['id'], 'type': conversation['conversation']['peer']['type'],
              'name': get_dialog_name(conversation, vk)}
    # формируем сообщения и вложения
    dialog['messages'], dialog['attachments'] = get_messages(vk, vk_tool, dialog['id'])
    return dialog


def get_dialog_name(conversation, vk):
    # если тип беседы - личный или группа,
    if conversation['conversation']['peer']['type'] in ['user', 'group']:
        # то добавляем пользователя в список {id: name}
        users_add(conversation['conversation']['peer']['id'], vk)
        # Получаем имя юзера или группы
        name = users[abs(conversation['conversation']['peer']['id'])]
    # если диалог - чат,
    elif conversation['conversation']['peer']['type'] == 'chat':
        # то получаем имя из title
        name = conversation['conversation']['chat_settings']['title']
    else:
        # иначе неожиданный результат, присвоим unknown
        name = r'{unknown}'
    # заменяем все "плохие" символы
    for c in ['\\', '/', ':', '*', '?', '<', '>', '|', '"']:
        name = name.replace(c, '_')
    return name


def users_add(uid, vk):
    try:
        # если имени нет в списке
        if uid not in users:
            # то если юзер (id>0)
            if uid > 0:
                # получаем инфо о человеке
                user = vk.users.get(user_ids=uid)[0]
                # если удален, то имя - deleted
                if ('deactivated' in user) and (user['deactivated'] == 'deleted') and (
                        user['first_name'] == 'DELETED'):
                    name = 'DELETED'
                else:
                    # если все норм - имя фамилия
                    name = f"{user['first_name']} {user['last_name']}"
            else:
                # если группа (ид <0), то получаем инфо об группе
                group = vk.messages.getConversationsById(peer_ids=uid, extended=1)['groups'][0]
                # и получаем имя
                name = group['name']
            # добавляем в список
            users[abs(uid)] = name
    except:
        # в случае ошибки присвоим неизвестное имя
        users[abs(uid)] = '{unknown user}'


# получаем сообщения и вложения
def get_messages(vk, vk_tools, cid):
    messages = []
    attachments = {}
    # читаем сообщения методом getHistory
    history = vk_tools.get_all(max_count=200, method='messages.getHistory',
                               values={'rev': 1, 'extended': 1, 'fields': 'first_name, last_name', 'peer_id': cid})
    # Для каждого сообщения
    for message in history['items']:
        # добавляем пользователя в список
        users_add(message['from_id'], vk)
        # обрабатываем сообщение
        mess = message_handler(message)
        # обрабатываем аттачменты
        attach = attach_handler(message)
        if mess: messages.append(mess)
        # добавляем в структура {type: [objects]}
        for at_type in attach:
            if at_type not in attachments: attachments[at_type] = []
            attachments[at_type] += attach[at_type]
    return messages, attachments


def message_handler(msg):
    # если сообщение не пустое,
    if len(msg['text']) > 0:
        text = [line for line in msg['text'].split('\n') if line]
        # то возвращаем текст, от кого, дату
        return {'text': text, 'from': users.get(msg['from_id']), 'date': msg['date']}
    return None


def attach_handler(msg):
    attachments = {}
    # для каждого вложения
    for attach in msg['attachments']:
        # определяем тип
        at_type = attach['type']
        at = attach[at_type]
        # если нужный тип не в списке вложений, то создаем список
        if at_type not in attachments and at_type in ['photo', 'doc', 'audio_message']:
            attachments[at_type] = []
        # в зависимости от типа, добавляем нужную структуру
        if at_type == 'photo':
            attachments[at_type].append(at['sizes'][-1]['url'])
        elif at_type == 'doc':
            attachments[at_type].append({'url': at['url'], 'access_key': at['access_key'], 'size': at['size'],
                                         'name': f"{at['title']}_{at['id']}.{at['ext']}", 'date': at['date']})
        elif at_type == 'audio_message':
            attachments[at_type].append({'url': at['link_ogg'], 'access_key': at['access_key'],
                                         'name': f"{at['owner_id']}_{at['id']}.oog"})
    return attachments


def dump_albums(albums, path):
    for album in albums:
        current_path = join(path, album['name'])
        makedirs(current_path, exist_ok=True)
        download_engine(album['photos'], current_path)


def dump_docs(docs, path):
    makedirs(path, exist_ok=True)
    download_engine(docs['docs'], path)


def dump_dialogs(dialogs, path):
    for dialog in dialogs:
        dialog_name = f"{dialog['name'].replace(' ', '_')}_{dialog['id']}"
        current_path = join(path, dialog_name)
        makedirs(current_path, exist_ok=True)
        with open(f'{join(current_path,dialog_name)}.json', 'w') as dialog_json:
            json.dump(dialog, dialog_json)
            for mode in dialog['attachments']:
                attachment_path = join(current_path, mode)
                download_engine(dialog['attachments'][mode], attachment_path)


def download_engine(object_list, path):
    makedirs(path, exist_ok=True)
    with Pool(4 * cpu_count()) as pool:
        pool.starmap(download, zip(object_list, repeat(path)))


def download(obj, root):
    try:
        filename, url = get_filename_url(obj)
        path = join(root, filename)
        if not exists(path):
            r = requests.get(url, stream=True, timeout=(30, 5))
            with open(path, 'wb') as f:
                shutil.copyfileobj(r.raw, f)
    except Exception as e:
        print(e)
        return False


def get_filename_url(obj):
    if not obj:
        raise ValueError('Объект пустой')
    if isinstance(obj, str):
        url = obj
        del obj
        filename = url.split('/')[-1]
    elif isinstance(obj, dict):
        url = obj.pop('url')
        options = obj
        filename, url = update_path(url, options=options)
    else:
        raise ValueError('Неизвестный объект')
    for char in ['\\', '/', ':', '*', '?', '<', '>', '|', '"']:
        filename = filename.replace(char, '_')
    return filename, url


def update_path(url, filename=None, options=None):
    if options:
        if 'name' in options:
            filename = '_'.join(options['name'].split(' '))
            if 'ext' in options:
                if filename.split('.')[-1] != options['ext']:
                    filename += f".{options['ext']}"
        if 'prefix' in options:
            filename = str(options['prefix']) + '_' + filename
        if 'access_key' in options:
            url = f"{url}?access_key={options['access_key']}"
    return filename, url


def main():
    token = '35e971d861baccf3847e5640866f838c981318e8834b30042cdbc2197a465173f39fd1dc8b3a445bae58f'
    vk = login_vk(token)
    vk_tools = vk_api.VkTools(vk)
    dump_albums(get_albums(vk, vk_tools), 'albums')
    dump_docs(get_docs(vk), 'docs')
    dump_dialogs(get_conversations(vk, vk_tools), 'dialogs')


if __name__ == '__main__':
    main()
